#!/usr/bin/python

################################################################
#                                                              #
# Persongen - A Norwegian SSN generator                        #
# written by Juan J. Guelfo @ Encripto AS                      #
# post@encripto.no                                             #
#                                                              #
# Copyright 2010-2015 Encripto AS. All rights reserved.        #
#                                                              #
# Persongen is licensed under the FreeBSD license.             #
# http://www.freebsd.org/copyright/freebsd-license.html        #
#                                                              #
################################################################



import sys, getopt, os, httplib, libs.out
from datetime import date

__version__ = "0.4"
__author__ = "Juan J. Guelfo, Encripto AS (post@encripto.no)"


def header():
    os.system("clear")
    print ""
    print " ================================================================= "
    print "|  Persongen v{0}: A Norwegian SSN generator\t\t\t  |".format(__version__)
    print "|  by {0}\t\t  |".format(__author__)
    print " ================================================================= "
    print ""


def help():
    header()
    print """
  Usage: python persongen.py [mandatory args]
 
  Mandatory args:
    -d date_of_birth        ...Generate a set of valid SSNs for a given date of birth (dd.mm.yyyy).
    -g gender               ...Generate SSNs based on gender (male, female or both).
    -t SSN_type             ...Generate SSNs based on type (resident, foreigner, both).
    -o output_file          ...Output file where the SSNs will be written to.

  Example:
    python persongen.py -d 01.12.1990 -g female -t resident -o ssn.txt

   """
    sys.exit()


def check_for_updates():
    try:
        http_conn = httplib.HTTPConnection("www.encripto.no", 80, timeout=10)
        headers = { "User-Agent" : "Mozilla/5.0 (compatible; Persongen/" + __version__ + ")"}
        http_conn.request("HEAD", "/tools/persongen-" + __version__ + ".tar.gz", None, headers) 
        http_resp = http_conn.getresponse()

        if http_resp.status == 404:
            print ""
            libs.out.print_ok("There is a new version of Persongen!")
            libs.out.print_ok("Check http://www.encripto.no/tools for more information.\n")

        elif http_resp.status == 200:
            libs.out.print_ok("You are running the latest version of Persongen.\n")

        http_conn.close()

    except:
        libs.out.print_error("Could not check for updates...\n")

    return



# Splitted date of birth validation
def validate_parameters(birth_year, birth_month, birth_day):

    try:
        check_date = date(birth_year, birth_month, birth_day)

    except ValueError:
        raise Exception("Please, enter a valid date in dd.mm.yyyy format.")


    if gender not in gender_types:
        raise Exception("Please, enter a valid gender (male, female, both).")

    if number_type not in number_types:
        raise Exception("Please, enter a valid SSN type (resident, foreigner, both).")

    if not output_file:
        raise Exception("Please, provide an output file name.")


# Calculates individual numbers based on gender, type and year
def calculate_individual_numbers(gender, gender_type, birth_year):

    individual_numbers = []

    #female
    if gender == gender_type[0]:
        if birth_year >= 1900 and birth_year <= 1999:
            individual_numbers = range(0, 499, 2)

        elif birth_year >= 1854 and birth_year <= 1899:
            individual_numbers = range(500, 749, 2)

        elif birth_year >= 2000 and birth_year <= 2039:
            individual_numbers = range(500, 999, 2)

        if birth_year >= 1940 and birth_year <= 1999:
            individual_numbers += range(900, 999, 2)

    #male
    elif gender == gender_type[1]:
        if birth_year >= 1900 and birth_year <= 1999:
            individual_numbers = range(1, 500, 2)

        elif birth_year >= 1854 and birth_year <= 1899:
            individual_numbers = range(501, 750, 2)

        elif birth_year >= 2000 and birth_year <= 2039:
            individual_numbers = range(501, 1000, 2)

        if birth_year >= 1940 and birth_year <= 1999:
            individual_numbers += range(901, 1000, 2)

    #both genders
    else:
        if birth_year >= 1900 and birth_year <= 1999:
            individual_numbers = range(0, 500)

        elif birth_year >= 1854 and birth_year <= 1899:
            individual_numbers = range(500, 750)

        elif birth_year >= 2000 and birth_year <= 2039:
            individual_numbers = range(500, 1000)

        if birth_year >= 1940 and birth_year <= 1999:
            individual_numbers += range(900, 1000)

    return individual_numbers


# Calculates SSNs based on date of birth, individual numbers, and SSN type
def calculate_SSNs(date_of_birth, individual_numbers, number_type, number_types):

    ind = ""
    SSNs = []
    for i in individual_numbers:
        ind = str(i).zfill(3)

        #SSNs for residents
        if number_type == number_types[0] or number_type == number_types[2]:
            c1 = 11 - ((3 * int(date_of_birth[0]) + 7 * int(date_of_birth[1]) 
                    + 6 * int(date_of_birth[3]) + 1 * int(date_of_birth[4]) 
                    + 8 * int(date_of_birth[8]) + 9 * int(date_of_birth[9]) 
                    + 4 * int(ind[0]) + 5 * int(ind[1]) 
                    + 2 * int(ind[2])) % 11)

            if c1 == 11:
                c1 = 0

            c2 = 11 - ((5 * int(date_of_birth[0]) + 4 * int(date_of_birth[1]) 
                    + 3 * int(date_of_birth[3]) + 2 * int(date_of_birth[4]) 
                    + 7 * int(date_of_birth[8]) + 6 * int(date_of_birth[9]) 
                    + 5 * int(ind[0]) + 4 * int(ind[1]) 
                    + 3 * int(ind[2]) + 2 * c1) % 11)

            if c2 == 11:
                c2 = 0

            if c1 != 10 and c2 != 10:
                SSNs.append(date_of_birth[0] + date_of_birth[1] + date_of_birth[3] 
                            + date_of_birth[4] + date_of_birth[8] + date_of_birth[9] 
                            + ind + str(c1) + str(c2))

        if number_type == number_types[1] or number_type == number_types[2]:
            c1 = 11 - ((3 * (int(date_of_birth[0]) + 4) + 7 * int(date_of_birth[1]) 
                    + 6 * int(date_of_birth[3]) + 1 * int(date_of_birth[4]) 
                    + 8 * int(date_of_birth[8]) + 9 * int(date_of_birth[9]) 
                    + 4 * int(ind[0]) + 5 * int(ind[1]) + 2 * int(ind[2])) % 11)

            if c1 == 11:
                c1 = 0

            c2 = 11 - ((5 * (int(date_of_birth[0]) + 4) + 4 * int(date_of_birth[1]) 
                    + 3 * int(date_of_birth[3]) + 2 * int(date_of_birth[4]) 
                    + 7 * int(date_of_birth[8]) + 6 * int(date_of_birth[9]) 
                    + 5 * int(ind[0]) + 4 * int(ind[1]) + 3 * int(ind[2]) + 2 * c1) % 11)

            if c2 == 11:
                c2 = 0

            if c1 != 10 and c2 != 10:
                SSNs.append(str(int(date_of_birth[0]) + 4) + date_of_birth[1] 
                            + date_of_birth[3] + date_of_birth[4] + date_of_birth[8] 
                            + date_of_birth[9] + ind + str(c1) + str(c2))

    return SSNs


# Write output to a file
def write_to_file(values, output_file):

    try:
        f = open(output_file, 'w')
        for v in values:
            f.write(v + "\n")

    except IOError:
        libs.out.print_error("Could not write SSNs to %s\n" % output_file)
        sys.exit()


if __name__ == '__main__':

    date_of_birth = ""
    gender = ""
    gender_types = ["female", "male", "both"]
    number_type = ""
    number_types = ["resident", "foreigner", "both"]
    output_file = ""

    # Parsing arguments
    try:
        options, args = getopt.getopt(sys.argv[1:], "d:g:t:o:", ["date=", "gender=", "type=", "output="])

    except getopt.GetoptError, err:
        header()
        libs.out.print_error("%s\n" % str(err))
        sys.exit(0)

    if not options:
        help()

    for opt, arg in options:
        if opt in ("-d"):
            date_of_birth = arg

        if opt in ("-g"):
            gender = arg

        if opt in ("-t"):
            number_type = arg

        if opt in ("-o"):
            output_file = arg


    #Input validation
    if len(date_of_birth) != 10:
        header()
        libs.out.print_error("Incorrect date format. Please, enter dd.mm.yyyy instead.\n")
        sys.exit()

    try:
        birth_day = int(date_of_birth[0] + date_of_birth[1])
        birth_month = int(date_of_birth[3] + date_of_birth[4])
        birth_year = int(date_of_birth[6] + date_of_birth[7] + date_of_birth[8] + date_of_birth[9])
        validate_parameters(birth_year, birth_month, birth_day)

    except ValueError:
        header()
        libs.out.print_error("Date of birth contains non-integer characters.\n")
        sys.exit()

    except Exception as e:
        header()
        libs.out.print_error("%s\n" % e)
        sys.exit()


    #Print header
    header()
    check_for_updates()

    libs.out.print_info("Calculating SSNs for date %s and %s gender(s)...\n" % (date_of_birth, gender))

    individual_numbers = calculate_individual_numbers(gender, gender_types, birth_year)
    SSNs = calculate_SSNs(date_of_birth, individual_numbers, number_type, number_types)
    write_to_file(SSNs, output_file)

    libs.out.print_ok("The results have been written to %s\n" % output_file)
    sys.exit()